/* scriptris - a free j2me tetris (R) clone with
 * bluetooth multiplayer support
 *
 *
 * (c) 2005-2006 Michael "ScriptKiller" Arndt
 * <scriptkiller@gmx.de>
 * http://scriptkiller.de/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */

import java.io.*;
import java.util.Vector;
import javax.microedition.lcdui.*;

/* jsr082 bluetooth api */
import javax.bluetooth.*;

import javax.microedition.io.*;

public class Bluetooth
    implements DiscoveryListener {

    Tetris t;
    String localDevName="";
    String serverName="";

    public static final String UUID="7219891290311a92282f0014a7082f09";
    public static final int SERVER_NAME_ID=0x5432;

    public BluetoothServer s=null;
    public BluetoothClient c=null;

    private LocalDevice local;

    /* server */
    private L2CAPConnectionNotifier server;
    private L2CAPConnection con;

    /* client */
    private DiscoveryAgent agent;
    /* RemoteDevice */
    private Vector devices=new Vector();

    /* ServiceRecord */
    private Vector records=new Vector();



    public Bluetooth(Tetris t) {
	this.t=t;

    }

    public boolean init() {

	try {
	    local=LocalDevice.getLocalDevice();
	    local.setDiscoverable(DiscoveryAgent.GIAC);
	    agent=local.getDiscoveryAgent();
	    localDevName=local.getFriendlyName();
	} catch(BluetoothStateException e) {
            System.err.println("Bluetooth Fehler: "+e);
	    return false;
	}	

	return true;

    }

    /* get bluetooth name of local device */
    public String getLocalDevName() {
	return localDevName;
    }

    /* set name of server in game */
    public void setServerName(String serverName) {
	this.serverName=serverName;
    }

    public boolean startServer() {

	s=new BluetoothServer(t, serverName);
	s.start();

	return true;
    }

    
    public void listen() {

    }

    public boolean startClient(Server s) {
	c=new BluetoothClient(t, s);
	c.start();
	return true;
    }


    /* stop server-thread */
    public void stopServer() {
	System.err.println("Stopping server ...");
	if(s!=null) {

	    s.send((byte)0x09);

	    try {
		/* wait a short period of time */
		s.sleep(500); /* milliseconds */
	    } catch(InterruptedException e) {
		System.err.println("Interrupted ..");
	    }

	    s.stop();
	}
    }

    /* send bye and stop client thread */
    public void stopClient() {
	System.err.println("Stopping client ...");

	/* still in inquiry? */
	if(agent!=null) {
	    agent.cancelInquiry(this);
	}
	/* TODO: cancel service searches? */

	if(c!=null) {
	    c.send((byte)0x09);

	    try {
		/* wait a short period of time */
		s.sleep(500); /* milliseconds */
	    } catch(InterruptedException e) {
		System.err.println("Interrupted ..");
	    }

	    c.stop();
	}
    }
    
    public void startInq() {
	/* clear vector */
        devices.removeAllElements();

	try {
            /* General/Unlimited Inquiry Access Code (GIAC) */
            agent.startInquiry(DiscoveryAgent.GIAC, this);
        } catch (BluetoothStateException e) {
            System.err.println("Kann Suche nicht starten: "+e);
        }
    }


    /* diese Methode wird vom System aufgerufen,
     * sobald die Suche nach Gerten abgeschlossen ist
     */
    public void inquiryCompleted(int discType) {


        switch(discType) {

        case INQUIRY_ERROR: {
            System.err.println("Fehler!\n");
            return;
        }
        case INQUIRY_TERMINATED: {
            devices.removeAllElements();
            System.err.println("Abgebrochen!\n");
            return;
        }
        case INQUIRY_COMPLETED: {
       if(devices.size() == 0) {
                System.err.println("Keine Gerte gefunden!\n");
		t.searchComplete(null);
            } else {
                System.err.println("OK, "+devices.size()+" Gert(e)\n");

                /* jetzt services suchen */
                //              startSearchServices();

		startServiceSearch();
            }
            return;
        }
        default: {
            System.err.println("Unbekannter discType: "+discType);
        }

        }



    }


    public void startServiceSearch() {
	System.out.println("startServiceSearch()");
	
	int attrSet[]=new int[1];
	attrSet[0]=Bluetooth.SERVER_NAME_ID;

        UUID[] searchList = new UUID[2];
	
        /*
         * Add the UUID for L2CAP to make sure that the service record
         * found will support L2CAP.  This value is defined in the
         * Bluetooth Assigned Numbers document.
         */
	searchList[0] = new UUID(0x0100);
	
        /*
         * Add the UUID for the printer service that we are going to use to
         * the list of UUIDs to search for. (a fictional printer service UUID)
         */
	searchList[1] = new UUID(this.UUID, false);

	/* vektor leeren */
        records.removeAllElements();
	
        /* fr jedes gefundene Gert */
        for(int i=0;i<devices.size();i++) {
	    
            RemoteDevice dev=(RemoteDevice)devices.elementAt(i);
            try {
                agent.searchServices(attrSet, searchList, dev, this);
            } catch(Exception e) {
                System.err.println("Kann keine Services suchen fr "+
                                   dev.getBluetoothAddress()+" ("+e+")");

                /* weitermachen in schleife */
                continue;
            }

        }



    }



    public void serviceSearchCompleted(int transID, int respCode) {

	System.out.println("serviceSearchCompleted() "+records.size());


	/* make servers array */
	Server servers[]=new Server[records.size()];

	for(int i=0;i<records.size();i++) {
	    ServiceRecord r=(ServiceRecord)records.elementAt(i);
	    String serverName="<noname>";
	    DataElement de=r.getAttributeValue(Bluetooth.SERVER_NAME_ID);

	    if(de!=null) {
		try {
		    serverName=(String)de.getValue();
		} catch(ClassCastException e) {
		}
	    }

	    servers[i]=new Server(r, serverName);
	}

	t.searchComplete(servers);


    }

    public void servicesDiscovered(int transID, ServiceRecord[] servRecord) {
	System.out.println("servicesDiscovered()");
	Form test=new Form("yay");

	/* alle gefundenen services zum Vektor hinzufgen */
        for(int i=0;i<servRecord.length;i++) {
            records.addElement(servRecord[i]);
	    //	    	    test.append(servRecord[i].getHostDevice().getBluetoothAddress());
	    //	    	    t.display.setCurrent(test);

        }
    }

    /* wird fr jedes gefundene Gert aufgerufen */
    public void deviceDiscovered(RemoteDevice btDevice, DeviceClass cod) {
	
        /* nur hinzufgen, wenn nicht schonmal gefunden,
         * sowas kann passieren ;-)
         */
        if(devices.indexOf(btDevice) == -1) {
            devices.addElement(btDevice);
            try {
                System.err.println(btDevice.getFriendlyName(false)+"\n");
            } catch(Exception e) {
            }

            System.err.println(btDevice.getBluetoothAddress()+"\n");
        }
    }


}


class BluetoothServer extends Thread {

    private Tetris t;
    L2CAPConnectionNotifier server;
    String UUID;
    String serverName;
    ServiceRecord record;
    private L2CAPConnection con;

    /* outbuf */
    byte buf[]=new byte[1];

    public BluetoothServer(Tetris t, String serverName) {
	this.t=t;
	this.UUID=Bluetooth.UUID;
	this.serverName=serverName;
    }

    public boolean send(byte b) {

	/* do we have an active connection? */
	if(con==null) {
	    return false;
	}


	buf[0]=b;
	try {
	    con.send(buf);
	} catch(IOException e) {
	    return false;
	}
	return true;
    }

    public void run() {
	
	try {
	    server=
		(L2CAPConnectionNotifier)
		Connector.open("btl2cap://localhost:"+UUID+";authorize=false;encrypt=false;name=Tetris");
	    System.err.println(server.toString());

	    /* todo: dont do this everytime server starts i guess +
	     * remove it when stopped */

	    /* add DataElement with server name */
	    record=LocalDevice.getLocalDevice().getRecord(server);
	    DataElement element=new DataElement(DataElement.STRING, serverName);
	    record.setAttributeValue(Bluetooth.SERVER_NAME_ID, element);

	    try {
		LocalDevice.getLocalDevice().updateRecord(record);
	    } catch(ServiceRegistrationException e) {

	    }

	} catch(SecurityException e) {
	    t.connectionError("Bluetooth Fehler: "+e);
	    System.err.println("Bluetooth Fehler: "+e);
	    return;
	} catch(BluetoothStateException e) {
	    t.connectionError("Bluetooth Fehler: "+e);
	    System.err.println("Bluetooth Fehler: "+e);
	    return;
	} catch(IOException e) {
	    t.connectionError("Bluetooth Fehler: "+e);
	    System.err.println("Bluetooth Fehler: "+e);
	    return;
	}

	
	try {
	    con=server.acceptAndOpen();

	    int length = con.getReceiveMTU();
	    byte data[] = new byte[length];

	    /* we have a connection */
	    t.clientConnected();


	    while(con.receive(data) != 0) {
		t.recvEvent(data[0]);
		System.out.println("Recv: "+data[0]);
	    }


	} catch(IOException e) {
	    System.err.println("IO Fehler: "+e);
	}

	return;
	
    }

    public void stop() {
	/* close connection if we are connected */
	if(con!=null) {
	    try {
		con.close();
	    } catch(IOException e) {
		System.err.println("IO Fehler: "+e);
	    }
	}

	if(server!=null) {
	    try {
		/* close server-notifier */
		server.close();
	    } catch(IOException e) {
		System.err.println("IO Fehler: "+e);
	    }
	}

    }

}

class BluetoothClient extends Thread {

    private Tetris t;
    private Server s;
    ServiceRecord rec;
    L2CAPConnection con;

    /* outbuf */
    byte buf[]=new byte[1];

    public BluetoothClient(Tetris t, Server s) {
	this.t=t;
	this.s=s;
    }
    
    public void run() {

	int mtu;

	try {
	    rec=s.record;
	    String url=rec.getConnectionURL(ServiceRecord.NOAUTHENTICATE_NOENCRYPT, false);


	    con=(L2CAPConnection)Connector.open(url);
            mtu=con.getTransmitMTU();

	    t.connectedToServer();

	} catch(Exception e) {
	    System.err.println("Kann nicht connecten: "+e);
	}

	try {
	    
	    while(true) {
		con.receive(buf);
		t.recvEvent(buf[0]);
		System.out.println("client recv: "+buf[0]);
	    }
	} catch(IOException e) {
	    System.err.println("IO Fehler: "+e);
	}



    }

    public void stop() {
	/* close connection if we are connected */
	if(con!=null) {
	    try {
		con.close();
	    } catch(IOException e) {
	    }
	}
    }


    public boolean send(byte b) {
	buf[0]=b;

	/* do we have an active connection? */
	if(con==null) {
	    return false;
	}

	try {
	    con.send(buf);
	} catch(IOException e) {
	    return false;
	}
	return true;
    }


}
