/* scriptris - a free j2me tetris (R) clone with
 * bluetooth multiplayer support
 *
 *
 * (c) 2005-2006 Michael "ScriptKiller" Arndt
 * <scriptkiller@gmx.de>
 * http://scriptkiller.de/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */

import java.io.*;
import javax.microedition.io.*;

public class TCP {

    Tetris t;
    TCPClient c;

    public TCP(Tetris t) {
	this.t=t;
    }

    public boolean init() {
	return true;
    }

    public void connect(String hostname, int port) {
	c=new TCPClient(t, hostname, port);
	c.start();
    }

    public void disconnect() {

	System.err.println("Disconnecting ...");
	if(c != null) {
	    c.stop();
	    c=null;
	}
    }

    public void sendNickName(String s) {

	byte size=(byte)s.length();
	byte bytes[]=s.getBytes();

	if(c == null) {
	    return;
	}

	c.send((byte)0x06);
	c.send(size);
	for(int i=0;i<size;i++) {
	    c.send(bytes[i]);
	}
    }
}

class TCPClient extends Thread {
    
    SocketConnection s;
    InputStream i;
    OutputStream o;
    Tetris t;

    private String hostname;
    private int port;

    /* outbuf */
    byte buf[]=new byte[1];
    
    public TCPClient(Tetris t, String hostname, int port) {
	this.t=t;
	this.port=port;
	this.hostname=hostname;
    }

    public void run() {
	
	byte b[]=new byte[1];

	t.showConnecting();

	try {

	    s=(SocketConnection)
		Connector.open("socket://"+hostname+":"+port);


	    System.err.println("Connected to server!");


	    i=s.openInputStream();
	    o=s.openOutputStream();

	    t.connectedToServer();

	    while(i.read(b) != -1) {
		System.err.println("HERE!");

		if(b[0] == 0x06) {
		    /* nickname */
		    i.read(b);
		    byte size=b[0];
		    byte buf[]=new byte[size];
		    for(int z=0;z<size;z++) {
			i.read(b);
			buf[z]=b[0];
		    }
		    String nickName=new String(buf);
		    t.nickNameReceived(nickName);
		    System.err.println("Nickname: "+nickName);
		}
		else {
		    t.recvEvent(b[0]);
		    System.err.println("Recv: "+b[0]);
		}
	    }

	    System.err.println("EOF");
	    t.connectionError("Serververbindung unterbrochen");


	}
	catch(Exception e) {
	    t.connectionError("TCP error :"+e);
	    System.err.println("TCP error: "+e);
	}

	return;
	
    }

    /* stop this thread */
    public void stop() {
	try {

	    if(s!=null) {
		s.close();
	    }
	    if(i!=null) {
		i.close();
	    }
	    if(o!=null) {
		o.close();
	    }

	    /* this causes i.read() to fail
	     * and thread to exit */

	} catch(IOException e) {
	    
	}
    }
    
    public boolean send(byte b) {
	System.err.println("Send "+b);
	buf[0]=b;
	try {
	    o.write(buf);
	} catch(IOException e) {
	    t.connectionError("Kann nicht senden: "+e);
	    return false;
	}
	return true;
    }

}
